use std::env;
use std::process;

fn main() {
    let args: Vec<String> = env::args().collect();
    if args.len() != 2 {
        println!("Error: You need to run this program with exactly 1 argument");
        process::exit(1);
    }
    let equation = computorv1::parse(&args[1]).unwrap_or_else(|e| {
        println!("Error during parsing: {e}");
        process::exit(1);
    });
    let is_equation = computorv1::maths::evaluator::is_equation(&equation);
    let evaluated = computorv1::maths::evaluator::evaluate(equation).unwrap_or_else(|e| {
        println!("Error during evaluation: {e}");
        process::exit(1);
    });
    if is_equation {
        println!("{} = 0", computorv1::format(&evaluated));
        computorv1::maths::solver::solve(evaluated);
    } else {
        println!("{} is equal to:", args[1]);
        println!("{}", computorv1::format(&evaluated));
    }
}
